"""
For now, this contains all the wrappers
"""

from functools import wraps
import os
from typing import Any
import time

from rich import print


def timeit(level: int = 1):
    """
    Prints the execution duration of the function.

    This is a decorator factory

    ## Args:
    - `level`: An arbitary int assigned to the wrapper. Use this combined with `WRAPWORKSLEVEL` env to control whether this wrapper uses print
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            start_time = time.perf_counter()
            if level >= int(os.getenv("WRAPWORKSLEVEL", "1")):
                    print(f"'{func.__name__}' starting")  # fmt:skip
            try:
                result = func(*args, **kwargs)
                return result
            finally:
                end_time = time.perf_counter()
                total_duration = end_time - start_time

                if level >= int(os.getenv("WRAPWORKSLEVEL", "1")):
                    print(f"'{func.__name__}' took {total_duration} seconds to execute!")  # fmt:skip

        return wrapper

    return decorator


def tryexcept(default_return: Any = None, level: int = 1):
    """
    Adds a try except block around the function saving to a bunch of keystrokes.

    This is a decorator factory.

    ## Args:
    - `default_return`: What to return when the function fails
    - `level`: An arbitary int assigned to the wrapper. Use this combined with `WRAPWORKSLEVEL` env to control whether this wrapper uses print
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                result = func(*args, **kwargs)
            except Exception as e:
                if level >= int(os.getenv("WRAPWORKSLEVEL", "1")):
                    print( f"Exception in function '{func.__name__}': {type(e).__name__} - {e}")  # fmt:skip
                return default_return
            return result

        return wrapper

    return decorator


def retry(max_retries=5, delay=1, default_return: Any = None, level=1):
    """
    Automatically retry the function.

    This is a decorator factory.

    ## Args:
    - `max_retries`: Maximum number of retires
    - `delay`: Delay in seconds between retries
    - `default_return`: What to return when the function fails
    - `level`: An arbitary int assigned to the wrapper. Use this combined with `WRAPWORKSLEVEL` env to control whether this wrapper uses print
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            retries = 0
            while retries < max_retries:
                try:
                    result = func(*args, **kwargs)
                    return result
                except Exception as e:
                    if level >= int(os.getenv("WRAPWORKSLEVEL", "1")):
                        print( f"Cxception in function '{func.__name__}': {type(e).__name__} - {e}")  # fmt:skip
                    retries += 1
                    time.sleep(delay)
                    continue
            if level >= int(os.getenv("WRAPWORKSLEVEL", "1")):
                print( f"Function '{func.__name__}' failed to execute successfully even after {retries} retries")  # fmt:skip
            return default_return

        return wrapper

    return decorator



"""
Unit tests for wraps
"""

# pylint:disable=wrong-import-position

import os
import sys
import unittest
from unittest.mock import patch
import time

sys.path.append(os.getcwd())
from wrapworks.wraps import timeit, tryexcept, retry


class TestTimeIt(unittest.TestCase):
    """Test Class for timeit"""

    def test_timeit_decorator(self):
        """Check if the time is being print"""

        @timeit()
        def test_function():
            time.sleep(1)

        with patch("wrapworks.wraps.print") as mock_print:
            test_function()

        self.assertEqual(mock_print.call_count, 2)


class TestTryExcept(unittest.TestCase):
    """Test Class for tryexcept"""

    def test_custom_return(self):
        """Test for custom return"""

        @tryexcept(default_return="Customer return")
        def test_function():
            raise ValueError()

        result = test_function()
        self.assertEqual(result, "Customer return")

    def test_good_return(self):
        """Test if the function return the child's return"""

        @tryexcept()
        def test_function():
            return 5

        result = test_function()
        self.assertEqual(result, 5)

    def test_print(self):
        """Test if the function prints the error"""

        @tryexcept()
        def test_function():
            raise ValueError()

        with patch("wrapworks.wraps.print") as mock_print:
            _ = test_function()

        mock_print.assert_called_once()


class TestRetryOnException(unittest.TestCase):
    """Test Class for retry"""

    def test_print_count(self):
        """Test for print count"""

        @retry(max_retries=3, delay=0)
        def test_function():
            raise ValueError()

        with patch("wrapworks.wraps.print") as mock_print:
            result = test_function()

        self.assertEqual(mock_print.call_count, 4)
        self.assertIsNone(result)


if __name__ == "__main__":
    unittest.main()



"""
Unit tests for error handling
"""

# pylint:disable=wrong-import-position

import os
import sys
import unittest
from unittest.mock import patch

sys.path.append(os.getcwd())
from wrapworks.error_handling import eprint


class TestEprintFunction(unittest.TestCase):
    """Test Class for eprint"""

    def test_eprint_with_traceback(self):
        """Test the function with traceback"""
        try:
            1 / 0
        except ZeroDivisionError as e:
            with patch("wrapworks.error_handling.print") as mock_print, patch(
                "traceback.print_tb"
            ) as mock_print_tb:
                eprint(e, self.test_eprint_with_traceback, no_tracebacks=False)

                mock_print.assert_called_once()
                mock_print_tb.assert_called_once()

    def test_eprint_without_traceback(self):
        """Test the function without traceback"""
        try:
            a = 1
            b = "2"
            c = a + b
        except TypeError as e:
            with patch("wrapworks.error_handling.print") as mock_print, patch(
                "traceback.print_tb"
            ) as mock_print_tb:
                eprint(e, self.test_eprint_without_traceback, no_tracebacks=True)

                mock_print.assert_called_once()
                mock_print_tb.assert_not_called()


if __name__ == "__main__":
    unittest.main()



"""
Contains helpers related to error handling
"""

from typing import Callable
import traceback

from rich import print


def eprint(e: Exception, func: Callable, no_tracebacks: bool = True):
    """
    Prints an error with traceback and type

    ## Args:
    - `e` : The exception
    - `func` : The funcation callable
    - `no_tracebacks`: Whether to include traceback
    """
    print(f"Error on '{func.__name__}'. {type(e).__name__} - {e}.")
    if not no_tracebacks:
        traceback.print_tb(e.__traceback__, limit=5)



"""
All helpers that deal with importing and exporting
"""

from typing import Any

import json
import pickle


def dump_json(path: str, data: dict):

    assert isinstance(
        data, dict
    ), f"Expected data to be of type dict, got {type(data).__name__} instead!"

    with open(path, "w", encoding="utf-8") as wf:
        json.dump(data, wf, default=str)


def load_json(path) -> dict:

    with open(path, "r", encoding="utf-8") as rf:
        data = json.load(rf)
        return data


def dump_text(path: str, data: str):

    assert isinstance(
        data, str
    ), f"Expected data to be of type str, got {type(data).__name__} instead!"
    with open(path, "w", encoding="utf-8") as wf:
        wf.write(data)
